from powa.dashboards import (
    Dashboard, Graph, Grid,
    MetricGroupDef, MetricDef,
    DashboardPage,
    ContentWidget)
from sqlalchemy.sql import (
    text, ColumnCollection, bindparam, table, select,
    literal_column, column, cast, case)
from sqlalchemy.types import Numeric
from powa.sql import qual_constants, resolve_quals, func
from powa.query import QueryOverview
from powa.sql.views import qualstat_getstatdata, qualstat_getstatdata


class QualConstantsMetricGroup(MetricGroupDef):
    """
    Metric group used for the qual charts.
    """
    name = "QualConstants"
    data_url = r"/metrics/database/(\w+)/query/(\w+)/qual/(\w+)/constants"
    xaxis = "rownumber"
    count = MetricDef(label="<%=group%>")
    grouper = "constants"

    @property
    def query(self):
        query = (qual_constants("most_executed",
                                text("""
            datname = :database AND
            s.queryid = :query AND
            qn.qualid = :qual AND
            coalesce_range && tstzrange(:from, :to)"""), top=10))
        base = qualstat_getstatdata()
        c = ColumnCollection(*base.inner_columns)
        base = base.where(c.queryid == bindparam("query")).alias()
        totals = (base.select()
                  .where((c.qualid == bindparam("qual")) &
                         (c.queryid == bindparam("query")))).alias()
        return (query.alias().select()
                .column(totals.c.count.label('total_count'))
                .column(base.c.queryid)
                .correlate(query))


    def post_process(self, data, database, query, qual, **kwargs):
        if not data['data']:
            return data
        conn = self.connect(database=database)
        max_rownumber = 0
        total_top10 = 0
        total = None
        d = {'total_count': 0}
        for d in data['data']:
            max_rownumber = max(max_rownumber, d['rownumber'])
            total_top10 += d['count']
        else:
            total = d['total_count']
        data['data'].append({'count': total - total_top10,
                     'rownumber': max_rownumber + 1,
                     'constants': 'Others'})
        return data


class QualDetail(ContentWidget):
    """
    Content widget showing detail for a specific qual.
    """
    title = "Detail for this Qual"
    data_url = r"/database/(\w+)/query/(\w+)/qual/(\w+)/detail"

    def get(self, database, query, qual):
        stmt = qualstat_getstatdata()
        c = ColumnCollection(*stmt.inner_columns)
        stmt = stmt.alias()
        stmt = (stmt.select()
            .where((c.qualid == bindparam("qualid")) &
                   (c.queryid== bindparam("query")))
            .where(stmt.c.count > 0)
            .column((c.queryid == bindparam("query")).label("is_my_query")))
        quals = list(self.execute(
            stmt,
            params={"query": query,
                    "from": self.get_argument("from"),
                    "to": self.get_argument("to"),
                    "qualid": qual}))
        quals = resolve_quals(self.connect(database=database), quals)
        my_qual = None
        other_queries = {}
        for qual in quals:
            if qual['is_my_query']:
                my_qual = qual
            else:
                other_queries[qual['queryid']] = qual['query']
        if my_qual is None:
            self.render("xhr.html", content="nodata")
            return
        self.render("database/query/qualdetail.html",
                    qual=my_qual,
                    database=database,
                    other_queries=other_queries)


class QualOverview(DashboardPage):
    """
    Dashboard page for a specific qual.
    """

    base_url = r"/database/(\w+)/query/(\w+)/qual/(\w+)"

    params = ["database", "query", "qual"]

    parent = QueryOverview

    datasources = [QualDetail, QualConstantsMetricGroup]

    dashboard = Dashboard(
        "Qual %(qual)s",
        [[QualDetail],
         [Graph("Most executed values",
               metrics=[QualConstantsMetricGroup.count],
               x_label_attr="constants",
               renderer="pie")]])



from sqlalchemy.sql import (select, cast, func, column, text, extract, case,
                            bindparam, literal_column, ColumnCollection)
from sqlalchemy.types import Numeric
from sqlalchemy.sql.functions import max, min, sum
from powa.sql.utils import *
from powa.sql import resolve_quals
from collections import defaultdict


class Biggest(object):

    def __init__(self, base_columns, order_by):
        self.base_columns = base_columns
        self.order_by = order_by

    def __call__(self, var, minval=0, label=None):
        label = label or var
        return func.greatest(
            func.lead(column(var))
            .over(order_by=self.order_by,
                  partition_by=self.base_columns)
                - column(var),
            minval).label(label)


def powa_base_statdata_detailed_db():
    base_query = text("""
    pg_database,
    LATERAL
    (
        SELECT unnested.dbid, unnested.queryid,(unnested.records).*
        FROM (
            SELECT psh.dbid, psh.queryid, psh.coalesce_range, unnest(records) AS records
            FROM powa_statements_history psh
            WHERE coalesce_range && tstzrange(:from, :to, '[]')
            AND psh.dbid = pg_database.oid
            AND psh.queryid IN ( SELECT powa_statements.queryid FROM powa_statements WHERE powa_statements.dbid = pg_database.oid )
        ) AS unnested
        WHERE tstzrange(:from, :to, '[]') @> (records).ts
        UNION ALL
        SELECT psc.dbid, psc.queryid,(psc.record).*
        FROM powa_statements_history_current psc
        WHERE tstzrange(:from,:to,'[]') @> (record).ts
        AND psc.dbid = pg_database.oid
        AND psc.queryid IN ( SELECT powa_statements.queryid FROM powa_statements WHERE powa_statements.dbid = pg_database.oid )
    ) h
    """)
    return base_query

def powa_base_statdata_db():
    base_query = text("""
    (
    SELECT pg_database.oid as dbid, h.*
    FROM
    pg_database LEFT JOIN
    (
          SELECT dbid, min(lower(coalesce_range)) AS min_ts, max(upper(coalesce_range)) AS max_ts
          FROM powa_statements_history_db dbh
          WHERE coalesce_range && tstzrange(:from, :to, '[]')
          GROUP BY dbid
    ) ranges ON pg_database.oid = ranges.dbid,
    LATERAL (
        SELECT (unnested1.records).*
        FROM (
            SELECT dbh.coalesce_range, unnest(records) AS records
            FROM powa_statements_history_db dbh
            WHERE coalesce_range @> min_ts
            AND dbh.dbid = ranges.dbid
        ) AS unnested1
        WHERE tstzrange(:from, :to, '[]') @> (unnested1.records).ts
        UNION ALL
        SELECT (unnested2.records).*
        FROM (
            SELECT dbh.coalesce_range, unnest(records) AS records
            FROM powa_statements_history_db dbh
            WHERE coalesce_range @> max_ts
            AND dbh.dbid = ranges.dbid
        ) AS unnested2
        WHERE tstzrange(:from, :to, '[]') @> (unnested2.records).ts
        UNION ALL
        SELECT (dbc.record).*
        FROM powa_statements_history_current_db dbc
        WHERE tstzrange(:from, :to, '[]') @> (dbc.record).ts
        AND dbc.dbid = pg_database.oid
    ) AS h) AS db_history
    """)
    return base_query

def get_diffs_forstatdata():
    return [
        diff("calls"),
        diff("total_time").label("runtime"),
        diff("shared_blks_read"),
        diff("shared_blks_hit"),
        diff("shared_blks_dirtied"),
        diff("shared_blks_written"),
        diff("temp_blks_read"),
        diff("temp_blks_written"),
        diff("blk_read_time"),
        diff("blk_write_time")
    ]

def powa_getstatdata_detailed_db():
    base_query = powa_base_statdata_detailed_db()
    diffs = get_diffs_forstatdata()
    return (select([
        column("queryid"),
        column("dbid"),
        column("datname"),
] + diffs)
        .select_from(base_query)
        .group_by(column("queryid"), column("dbid"), column("datname"))
        .having(max(column("calls")) - min(column("calls")) > 0))

def powa_getstatdata_db():
    base_query = powa_base_statdata_db()
    diffs = get_diffs_forstatdata()
    return (select([column("dbid")] + diffs)
            .select_from(base_query)
            .group_by(column("dbid"))
            .having(max(column("calls")) - min(column("calls")) > 0))


BASE_QUERY_SAMPLE_DB = text("""(
    SELECT datname, base.* FROM pg_database,
    LATERAL (
        SELECT *
        FROM (
            SELECT
            row_number() OVER (PARTITION BY dbid ORDER BY statements_history.ts) AS number,
            count(*) OVER (PARTITION BY dbid) AS total,
            *
            FROM (
                SELECT dbid, (unnested.records).*
                FROM (
                    SELECT psh.dbid, psh.coalesce_range, unnest(records) AS records
                    FROM powa_statements_history_db psh
                    WHERE coalesce_range && tstzrange(:from, :to,'[]')
                    AND psh.dbid = pg_database.oid
                ) AS unnested
                WHERE tstzrange(:from, :to, '[]') @> (records).ts
                UNION ALL
                SELECT dbid, (record).*
                FROM powa_statements_history_current_db
                WHERE tstzrange(:from, :to, '[]') @> (record).ts
                AND dbid = pg_database.oid
            ) AS statements_history
        ) AS sh
        WHERE number % ( int8larger((total)/(:samples+1),1) ) = 0
    ) AS base
) AS by_db
""")

BASE_QUERY_SAMPLE = text("""(
    SELECT datname, dbid, queryid, base.*
    FROM powa_statements JOIN pg_database ON pg_database.oid = powa_statements.dbid,
    LATERAL (
        SELECT *
        FROM (SELECT
            row_number() OVER (PARTITION BY queryid ORDER BY statements_history.ts) AS number,
            count(*) OVER (PARTITION BY queryid) AS total,
            *
            FROM (
                SELECT (unnested.records).*
                FROM (
                    SELECT psh.queryid, psh.coalesce_range, unnest(records) AS records
                    FROM powa_statements_history psh
                    WHERE coalesce_range && tstzrange(:from, :to, '[]')
                    AND psh.queryid = powa_statements.queryid
                ) AS unnested
                WHERE tstzrange(:from, :to, '[]') @> (records).ts
                UNION ALL
                SELECT (record).*
                FROM powa_statements_history_current phc
                WHERE tstzrange(:from, :to, '[]') @> (record).ts
                AND phc.queryid = powa_statements.queryid
            ) AS statements_history
        ) AS sh
        WHERE number % ( int8larger((total)/(:samples+1),1) ) = 0
    ) AS base
) AS by_query
""")


def powa_getstatdata_sample(mode):
    if mode == "db":
        base_query = BASE_QUERY_SAMPLE_DB
        base_columns = ["dbid"]

    elif mode == "query":
        base_query = BASE_QUERY_SAMPLE
        base_columns = ["dbid", "queryid"]

    ts = column('ts')
    biggest = Biggest(base_columns, ts)


    return select(base_columns + [
        ts,
        biggest("ts", '0 s', "mesure_interval"),
        biggest("calls"),
        biggest("total_time", label="runtime"),
        biggest("rows"),
        biggest("shared_blks_read"),
        biggest("shared_blks_hit"),
        biggest("shared_blks_dirtied"),
        biggest("shared_blks_written"),
        biggest("local_blks_read"),
        biggest("local_blks_hit"),
        biggest("local_blks_dirtied"),
        biggest("local_blks_written"),
        biggest("temp_blks_read"),
        biggest("temp_blks_written"),
        biggest("blk_read_time"),
        biggest("blk_write_time")]).select_from(base_query).apply_labels()


def qualstat_base_statdata():
    base_query = text("""
    (
    SELECT queryid, qualid, (unnested.records).*
    FROM (
        SELECT pqnh.qualid, pqnh.queryid, pqnh.dbid, pqnh.userid, pqnh.coalesce_range, unnest(records) as records
        FROM powa_qualstats_quals_history pqnh
        WHERE coalesce_range  && tstzrange(:from, :to, '[]') AND pqnh.queryid = :query
    ) AS unnested
    WHERE tstzrange(:from, :to, '[]') @> (records).ts
    UNION ALL
    SELECT queryid, qualid, pqnc.ts, pqnc.count, pqnc.nbfiltered
    FROM powa_qualstats_quals_history_current pqnc
    WHERE tstzrange(:from, :to, '[]') @> pqnc.ts AND pqnc.queryid = :query
    ) h
    JOIN powa_qualstats_quals pqnh USING (queryid, qualid)
    """)
    return base_query


def qualstat_getstatdata():
    base_query = qualstat_base_statdata()
    return (select([
        column("qualid"),
        column("queryid").label("queryid"),
        func.to_json(column("quals")).label("quals"),
        sum(column("count")).label("count"),
        sum(column("nbfiltered")).label("nbfiltered"),
        case(
            [(sum(column("count")) == 0, 0)],
            else_=sum(column("nbfiltered")) /
                cast(sum(column("count")), Numeric)
        ).label("filter_ratio")])
        .select_from(base_query)
        .group_by(column("qualid"), literal_column("queryid"), column("quals")))

def possible_indexes(resolved_qual_list):
    by_am = defaultdict(list)
    for qual in resolved_qual_list:
        for am in qual['indexam_names']:
            by_am[am].append(qual)
    return by_am


BASE_QUERY_KCACHE_SAMPLE = text("""
        powa_statements s JOIN pg_database ON pg_database.oid = s.dbid,
        LATERAL (
            SELECT *
            FROM (
                SELECT row_number() OVER (ORDER BY kmbq.ts) AS number,
                    count(*) OVER () as total,
                        *
                FROM (
                    SELECT km.ts,
                    sum(km.reads) AS reads, sum(km.writes) AS writes,
                    sum(km.user_time) AS user_time, sum(km.system_time) AS system_time
                    FROM (
                        SELECT * FROM (
                            SELECT (unnest(metrics)).*
                            FROM powa_kcache_metrics km
                            WHERE km.queryid = s.queryid
                            AND km.coalesce_range && tstzrange(:from, :to, '[]')
                        ) his
                        WHERE tstzrange(:from, :to, '[]') @> his.ts
                        UNION ALL
                        SELECT (metrics).*
                        FROM powa_kcache_metrics_current kmc
                        WHERE tstzrange(:from, :to, '[]') @> (metrics).ts
                        AND kmc.queryid = s.queryid
                    ) km
                    GROUP BY km.ts
                ) kmbq
            ) kmn
        WHERE kmn.number % (int8larger(total/(:samples+1),1) ) = 0
        ) kcache
""")

def kcache_getstatdata_sample():
    base_query = BASE_QUERY_KCACHE_SAMPLE
    base_columns = ["queryid", "datname"]
    ts = column('ts')
    biggest = Biggest(base_columns, ts)

    return (select(base_columns + [
        ts,
        biggest("reads"),
        biggest("writes"),
        biggest("user_time"),
        biggest("system_time")])
        .select_from(base_query)
        .apply_labels())



from tornado.web import UIModule

class UserMenu(UIModule):

    def render(self):
        return ""


class SideBar(UIModule):

    def render(self, menu):
        if self.current_user:
            return self.render_string("sidebar.html",
                                      menu=menu,
                                      databases=self.handler.databases,
                                      current_database=self.handler.database)
        else:
            return ""


class Messages(UIModule):

    def render(self):
        return self.render_string("messages.html")


class MenuEntry(UIModule):

    def __init__(self, title, url_name, url_params=None, children=None, active=False):
        self.title = title
        self.url_name = url_name
        self.children = children or []
        self.url_params = url_params or {}
        self.active = active

    def findMenu(self, url_name, url_params):
        if self.url_name == url_name:
            if all(url_params.get(key) == value
                   for key, value in self.url_params.items()):
                return self
        for child in self.children:
            result = child.findMenu(url_name, url_params)
            if result is not None:
                return result
        return None

    def render(self, handler):
        return handler.render_string("menuitem.html",
                           menu=self)

    def get_breadcrumb(self):
        base = []
        if self.active and self.url_name:
            base.append(self)
        for child in self.children:
            base.extend(child.get_breadcrumb())
        return base

GlobalMenu = [MenuEntry("Server configuration", "ConfigOverview"),
              MenuEntry("All databases", "Overview")]



"""
Utilities for the basis of Powa
"""
from tornado.web import RequestHandler, authenticated, HTTPError
from powa import ui_methods
from powa.json import to_json
from sqlalchemy import create_engine, text
from sqlalchemy.engine.url import URL
from tornado.options import options
import pickle


class BaseHandler(RequestHandler):
    """
    Subclass of Tornado RequestHandler adding a bunch
    of utility methods.
    """

    def __init__(self, *args, **kwargs):
        super(BaseHandler, self).__init__(*args, **kwargs)
        self.flashed_messages = {}
        self._databases = None
        self._connections = {}

    def render_json(self, value):
        """
        Render the object as json response.
        """
        self.set_header('Content-Type', 'application/json')
        self.write(to_json(value))

    @property
    def current_user(self):
        """
        Return the current_user if he is allowed to connect
        to his server of choice.
        """
        raw = self.get_secure_cookie('username')
        if raw:
            user = raw.decode('utf8')
            try:
                self.connect()
                return user
            except Exception as _:
                return None

    @property
    def menu(self):
        return None

    @property
    def database(self):
        """Return the current database."""
        return None

    @property
    def databases(self, **kwargs):
        """
        Return the list of databases in this instance.
        """
        if self.current_user:
            if self._databases is None:
                self._databases = [d[0] for d in self.execute(
                    "SELECT datname FROM pg_database WHERE datallowconn ORDER BY DATNAME",
                    **kwargs)]
            return self._databases

    def on_finish(self):
        for engine in self._connections.values():
            engine.dispose()

    def connect(self, server=None, username=None, password=None,
                database=None):
        """
        Connect to a specific database.
        Parameters default values are taken from the cookies and the server
        configuration file.
        """
        server = server or self.get_secure_cookie('server').decode('utf8')
        username = username or self.get_secure_cookie('username').decode('utf8')
        password = (password or
                    self.get_secure_cookie('password').decode('utf8'))
        if server not in options.servers:
            raise HTTPError(404)
        connoptions = options.servers[server].copy()
        connoptions['username'] = username
        connoptions['password'] = password
        if database is not None:
            connoptions['database'] = database
        engineoptions = {'_initialize': False}
        if self.application.settings['debug']:
            engineoptions['echo'] = True
        url = URL("postgresql+psycopg2", **connoptions)
        if url in self._connections:
            return self._connections.get(url)
        engine = create_engine(url, **engineoptions)
        engine.connect()
        self._connections[url] = engine
        return engine

    def has_extension(self, extname):
        """
        Returns whether the database has the specific extension installed.
        """
        return self.execute(text(
            """
            SELECT true FROM pg_extension WHERE extname = :extname
            """), params={"extname": extname}).rowcount > 0

    def write_error(self, status_code, **kwargs):
        if status_code == 403:
            self._status_code = status_code
            self.clear_all_cookies()
            self.flash("Authentification failed", "alert")
            self.render("login.html", title="Login")
            return
        if status_code == 501:
            self._status_code = status_code
            self.render("xhr.html", content=kwargs["exc_info"][1].log_message)
            return
        super(BaseHandler, self).write_error(status_code, **kwargs)

    def execute(self, query, params=None, server=None, username=None,
                database=None,
                password=None):
        """
        Execute a query against a database, with specific bind parameters.
        """
        if params is None:
            params = {}
        engine = self.connect(server, username, password, database)
        return engine.execute(query, **params)

    def get_pickle_cookie(self, name):
        """
        Deserialize a cookie value using the pickle protocol.
        """
        value = self.get_secure_cookie(name)
        if value:
            return pickle.loads(value)

    def set_pickle_cookie(self, name, value):
        """
        Serialize a cookie value using the pickle protocol.
        """
        self.set_secure_cookie(name, pickle.dumps(value))

    flash = ui_methods.flash
    reverse_url_with_params = ui_methods.reverse_url_with_params


class AuthHandler(BaseHandler):
    """
    Base handler for urls needing authentifications.
    """

    @authenticated
    def prepare(self):
        super(AuthHandler, self).prepare()



"""
Module containing the by-database dashboard.
"""
from powa.framework import AuthHandler
from powa.dashboards import (
    Dashboard, Graph, Grid,
    MetricGroupDef, MetricDef,
    DashboardPage)

from powa.sql.views import (block_size,
                            powa_getstatdata_detailed_db, mulblock,
                            powa_getstatdata_sample, total_read, total_hit,
                            total_measure_interval,
                            to_epoch)
from powa.overview import Overview
from sqlalchemy.sql import ColumnCollection, bindparam, column, select
from sqlalchemy.sql.functions import sum
from powa.sql.utils import greatest
from powa.sql.tables import powa_statements


class DatabaseSelector(AuthHandler):
    """Page allowing to choose a database."""

    def get(self):
        self.redirect(self.reverse_url(
            'DatabaseOverview',
            self.get_argument("database")))


class DatabaseOverviewMetricGroup(MetricGroupDef):
    """Metric group for the database global graphs."""
    name = "database_overview"
    xaxis = "ts"
    data_url = r"/metrics/database_overview/(\w+)/"
    avg_runtime = MetricDef(label="Avg runtime", type="duration")
    total_blks_hit = MetricDef(label="Total hit", type="sizerate")
    total_blks_read = MetricDef(label="Total read", type="sizerate")

    @property
    def query(self):
        # Fetch the base query for sample, and filter them on the database
        bs = block_size.c.block_size
        subquery = powa_getstatdata_sample("db")
        # Put the where clause inside the subquery
        subquery = subquery.where(column("datname") == bindparam("database"))
        query = subquery.alias()
        c = query.c
        return (select([
                to_epoch(c.ts),
                (sum(c.runtime) / greatest(sum(c.calls), 1.)).label("avg_runtime"),
                total_read(c),
                total_hit(c)])
            .where(c.calls != None)
            .group_by(c.ts, bs)
            .order_by(c.ts)
            .params(samples=100))



class ByQueryMetricGroup(MetricGroupDef):
    """Metric group for indivual query stats (displayed on the grid)."""
    name = "all_queries"
    xaxis = "queryid"
    axis_type = "category"
    data_url = r"/metrics/database_all_queries/(\w+)/"
    calls = MetricDef(label="#", type="string")
    runtime = MetricDef(label="Time", type="duration")
    avg_runtime = MetricDef(label="Avg time", type="duration")
    blk_read_time = MetricDef(label="Read", type="duration")
    blk_write_time = MetricDef(label="Write", type="duration")
    shared_blks_read = MetricDef(label="Read", type="size")
    shared_blks_hit = MetricDef(label="Hit", type="size")
    shared_blks_dirtied = MetricDef(label="Dirtied", type="size")
    shared_blks_written = MetricDef(label="Written", type="size")
    temp_blks_read = MetricDef(label="Read", type="size")
    temp_blks_written = MetricDef(label="Written", type="size")

    # TODO: refactor with GlobalDatabasesMetricGroup
    @property
    def query(self):
        # Working from the statdata detailed_db base query
        inner_query = powa_getstatdata_detailed_db()
        inner_query = inner_query.alias()
        c = inner_query.c
        ps = powa_statements
        # Multiply each measure by the size of one block.
        columns = [c.queryid,
                   ps.c.query,
                   c.calls,
                   c.runtime,
                   mulblock(c.shared_blks_read),
                   mulblock(c.shared_blks_hit),
                   mulblock(c.shared_blks_dirtied),
                   mulblock(c.shared_blks_written),
                   mulblock(c.temp_blks_read),
                   mulblock(c.temp_blks_written),
                   (c.runtime / greatest(c.calls, 1)).label("avg_runtime"),
                   c.blk_read_time,
                   c.blk_write_time]
        from_clause = inner_query.join(ps,
                                       (ps.c.queryid == c.queryid) &
                                       (ps.c.dbid == c.dbid))
        return (select(columns)
                .select_from(from_clause)
                .where(c.datname == bindparam("database"))
                .order_by(c.calls.desc()))


    def process(self, val, database=None, **kwargs):
        val = dict(val)
        val["url"] = self.reverse_url(
            "QueryOverview", database, val["queryid"])
        return val


class DatabaseOverview(DashboardPage):
    """DatabaseOverview Dashboard."""
    base_url = r"/database/(\w+)/overview"
    datasources = [DatabaseOverviewMetricGroup, ByQueryMetricGroup]
    params = ["database"]
    parent = Overview
    dashboard = Dashboard(
        "Database overview for %(database)s",
        [[Graph("Calls (On database %(database)s)",
                metrics=[DatabaseOverviewMetricGroup.avg_runtime]),
          Graph("Blocks (On database %(database)s)",
                metrics=[DatabaseOverviewMetricGroup.total_blks_read,
                         DatabaseOverviewMetricGroup.total_blks_hit])],
         [Grid("Details for all queries",
                toprow=[{
                    'merge': True
                },{
                    'name': 'Execution',
                    'merge': False,
                    'colspan': 3
                }, {
                    'name': 'I/O Time',
                    'merge': False,
                    'colspan': 2
                }, {
                    'name': 'Blocks',
                    'merge': False,
                    'colspan': 4,
                }, {
                    'name': 'Temp blocks',
                    'merge': False,
                    'colspan': 2
                }],
               columns=[{
                   "name": "query",
                   "label": "Query",
                   "type": "query",
                   "url_attr": "url",
                   "max_length": 70
               }],
               metrics=ByQueryMetricGroup.all())]])

    @classmethod
    def get_menutitle(cls, handler, params):
        return params.get("database")



from powa.dashboards import (
    Dashboard, Grid,
    MetricGroupDef, MetricDef,
    DashboardPage)

class PgSettingsMetricGroup(MetricGroupDef):
    """
    Metric group for the pg_settings grid.
    """

    name = "pg_settings"
    xaxis = "setting_name"
    data_url = r"/config/pg_settings/"
    axis_type = "category"
    setting_value = MetricDef(label="Value", type="string")
    setting_unit = MetricDef(label="Unit", type="string")
    query = """
            SELECT name as setting_name, setting as setting_value, COALESCE(unit,'') AS setting_unit
            FROM pg_settings
            --WHERE name like 'powa%%'
            ORDER BY name"""

class PgExtensionsMetricGroup(MetricGroupDef):
    """
    Metric group for the pg_settings grid.
    """

    name = "pg_extensions"
    xaxis = "extname"
    data_url = r"/config/pg_extensions/"
    axis_type = "category"
    available = MetricDef(label="Extension available", type="bool")
    installed = MetricDef(label="Extension installed", type="bool")
    handled = MetricDef(label="Extension handled", type="bool")
    query = """
           SELECT DISTINCT s.extname,
             CASE WHEN avail.name IS NULL then false ELSE true END AS available,
             CASE WHEN ins.extname IS NULL then false ELSE true END AS installed,
             CASE WHEN f.module IS NULL then false ELSE true END AS handled
           FROM (SELECT 'pg_stat_statements' AS extname UNION SELECT 'pg_qualstats' UNION SELECT 'pg_stat_kcache') s
           LEFT JOIN pg_available_extensions avail on s.extname = avail.name
           LEFT JOIN pg_extension ins on s.extname = ins.extname
           LEFT JOIN powa_functions f ON s.extname = f.module
            """


class ConfigOverview(DashboardPage):
    """
    Dashboard page for configuration page.
    """

    base_url = r"/config/"

    datasources = [PgSettingsMetricGroup, PgExtensionsMetricGroup]

    dashboard = Dashboard(
        "Configuration overview",
         [[Grid("Extensions",
               columns=[{
                   "name": "extname",
                   "label": "Extensions",
                   "url_attr": "url"
               }],
               metrics=PgExtensionsMetricGroup.all()
          ),
          Grid("PostgreSQL settings",
               columns=[{
                   "name": "setting_name",
                   "label": "Setting",
                   "url_attr": "url"
               }],
               metrics=PgSettingsMetricGroup.all()
               )
         ]]
    )

    @classmethod
    def get_menutitle(cls, handler, params):
        return "Configuration overview"



from sqlalchemy.sql import table, column

powa_statements = table("powa_statements",
                        column("query"),
                        column("queryid"),
                        column("dbid"),
                        column("userid"))

pg_database = table(
    "pg_database",
    column("oid"),
    column("datname"))




from powa.dashboards import MetricGroupDef, MetricDef

class Detail(MetricGroupDef):
    calls = MetricDef(label="#Calls", type="string")
    runtime = MetricDef(label="Runtime", type="duration")
    avg_runtime = MetricDef(label="Avg runtime", type="duration")
    shared_blks_read = MetricDef(label="Blocks read", type="size")
    shared_blks_hit = MetricDef(label="Blocks hit", type="size")
    shared_blks_dirtied = MetricDef(label="Blocks dirtied", type="size")
    shared_blks_written = MetricDef(label="Blocks written", type="size")
    temp_blks_written = MetricDef(label="Temp Blocks written", type="size")


class Totals(MetricGroupDef):
    avg_runtime = MetricDef(label="Total runtime", type="duration")
    total_blks_hit = MetricDef(label="Total hit", type="sizerate")
    total_blks_read = MetricDef(label="Total read", type="sizerate")



"""
Index page presenting an overview of the cluster stats.
"""

from powa.dashboards import (
    Dashboard, Graph, Grid,
    MetricGroupDef, MetricDef,
    DashboardPage)

from powa.sql.views import (
    block_size, powa_getstatdata_db,
    powa_getstatdata_sample)
from sqlalchemy.sql.functions import sum
from sqlalchemy.sql import select, cast, extract
from sqlalchemy.types import Numeric
from powa.sql.utils import total_read, total_hit, mulblock, round, greatest
from powa.sql.tables import pg_database


class ByDatabaseMetricGroup(MetricGroupDef):
    """
    Metric group used by the "by database" grid
    """
    name = "by_database"
    xaxis = "datname"
    data_url = r"/metrics/by_databases/"
    axis_type = "category"
    calls = MetricDef(label="#Calls", type="string")
    runtime = MetricDef(label="Runtime", type="duration")
    avg_runtime = MetricDef(label="Avg runtime", type="duration")
    shared_blks_read = MetricDef(label="Blocks read", type="size")
    shared_blks_hit = MetricDef(label="Blocks hit", type="size")
    shared_blks_dirtied = MetricDef(label="Blocks dirtied", type="size")
    shared_blks_written = MetricDef(label="Blocks written", type="size")
    temp_blks_written = MetricDef(label="Temp Blocks written", type="size")
    io_time = MetricDef(label="I/O time")

    @property
    def query(self):
        bs = block_size.c.block_size
        inner_query = powa_getstatdata_db().alias()
        c = inner_query.c
        from_clause = inner_query.join(
            pg_database,
            c.dbid == pg_database.c.oid)

        return (select([
            pg_database.c.datname,
            sum(c.calls).label("calls"),
            sum(c.runtime).label("runtime"),
            round(cast(sum(c.runtime), Numeric) / greatest(sum(c.calls), 1), 2).label("avg_runtime"),
            mulblock(sum(c.shared_blks_read).label("shared_blks_read")),
            mulblock(sum(c.shared_blks_hit).label("shared_blks_hit")),
            mulblock(sum(c.shared_blks_dirtied).label("shared_blks_dirtied")),
            mulblock(sum(c.shared_blks_written).label("shared_blks_written")),
            mulblock(sum(c.temp_blks_written).label("temp_blks_written")),
            round(cast(sum(c.blk_read_time + c.blk_write_time), Numeric), 2).label("io_time")
        ])
            .select_from(from_clause)
            .order_by(sum(c.calls).desc())
            .group_by(pg_database.c.datname, bs))


    def process(self, val, **kwargs):
        val = dict(val)
        val["url"] = self.reverse_url("DatabaseOverview", val["datname"])
        return val


class GlobalDatabasesMetricGroup(MetricGroupDef):
    """
    Metric group used by summarized graphs.
    """
    name = "all_databases"
    data_url = r"/metrics/databases_globals/"
    avg_runtime = MetricDef(label="Total runtime", type="duration")
    total_blks_hit = MetricDef(label="Total hit", type="sizerate")
    total_blks_read = MetricDef(label="Total read", type="sizerate")

    @property
    def query(self):
        bs = block_size.c.block_size
        query = powa_getstatdata_sample("db")
        query = query.alias()
        c = query.c
        return (select([
                extract("epoch", c.ts).label("ts"),
                (sum(c.runtime) / greatest(sum(c.calls), 1)).label("avg_runtime"),
                total_read(c),
                total_hit(c)])
            .where(c.calls != None)
            .group_by(c.ts, bs)
            .order_by(c.ts)
            .params(samples=100))



class Overview(DashboardPage):
    """
    Overview dashboard page.
    """
    base_url = r"/overview/"
    datasources = [GlobalDatabasesMetricGroup, ByDatabaseMetricGroup]
    dashboard = Dashboard(
        "All databases",
        [[Graph("Query runtime per second (all databases)",
                metrics=[GlobalDatabasesMetricGroup.avg_runtime]),
          Graph("Block access in Bps",
                metrics=[GlobalDatabasesMetricGroup.total_blks_hit,
                         GlobalDatabasesMetricGroup.total_blks_read])],
         [Grid("Details for all databases",
               columns=[{
                   "name": "datname",
                   "label": "Database",
                   "url_attr": "url"
               }],
               metrics=ByDatabaseMetricGroup.all())]])

    @classmethod
    def get_menutitle(cls, handler, params):
        return "All databases"

    @classmethod
    def get_childmenu(cls, handler, params):
        from powa.database import DatabaseOverview
        children = []
        for d in list(handler.databases):
            new_params = params.copy()
            new_params["database"] = d
            children.append(DatabaseOverview.get_selfmenu(handler, new_params))
        return children



from __future__ import print_function
"""
Powa main application.
"""

__VERSION__ = '0.0.1'

from tornado.web import Application, URLSpec as U
from tornado.options import define, parse_config_file, options
from powa import ui_modules, ui_methods
from powa.framework import AuthHandler
from powa.user import LoginHandler, LogoutHandler
from powa.overview import Overview
from powa.database import DatabaseSelector, DatabaseOverview
from powa.query import QueryOverview
from powa.qual import QualOverview
from powa.config import ConfigOverview
import os
import sys

class IndexHandler(AuthHandler):
    """
    Handler for the main page.
    """

    def get(self):
        return self.redirect("/overview/")


URLS = [
    U(r"/login/", LoginHandler, name="login"),
    U(r"/logout/", LogoutHandler, name="logout"),
    U(r"/database/select", DatabaseSelector, name="database_selector"),
    U(r"/", IndexHandler, name="index")
]



for dashboard in (Overview,
                  DatabaseOverview,
                  QueryOverview,
                  QualOverview,
                  ConfigOverview):
    URLS.extend(dashboard.url_specs())


POWA_ROOT = os.path.dirname(__file__)
CONF_LOCATIONS=['/etc/powa.conf',
        os.path.expanduser('~/.config/powa.conf'),
        os.path.expanduser('~/.powa.conf'),
        'powa.conf']


SAMPLE_CONFIG_FILE="""
servers={
  'main': {
    'host': 'localhost',
    'port': '5432',
    'database': 'powa'
  }
}
cookie_secret="SUPERSECRET_THAT_YOU_SHOULD_CHANGE"
"""

def make_app(**kwargs):
    """
    Parse the config file and instantiate a tornado app.
    """
    define("servers", type=dict)
    define("cookie_secret", type=str)
    for possible_config in CONF_LOCATIONS:
        try:
            parse_config_file(possible_config)
        except:
            pass
    for key in ('servers', 'cookie_secret'):
        if getattr(options, key, None) is None:
            print("You should define a server and cookie_secret in your configuration file.")
            print("""Place and adapt the following content in one of those locations:""")
            print("\n\t".join([""] + CONF_LOCATIONS))
            print(SAMPLE_CONFIG_FILE)
            sys.exit(-1)

    return Application(
        URLS,
        ui_modules=ui_modules,
        ui_methods=ui_methods,
        login_url="/login/",
        static_path=os.path.join(POWA_ROOT, "static"),
        cookie_secret=options.cookie_secret,
        template_path=os.path.join(POWA_ROOT,  "templates"),
        **kwargs)



from setuptools import setup, find_packages

__VERSION__ = None

with open('powa/__init__.py') as f:
    for line in f:
        if line.startswith('__VERSION__'):
            __VERSION__ = line.split('=')[1].replace("'", '').strip()


requires = ['sqlalchemy', 'tornado', 'psycopg2'],

setup(
    name='powa-web',
    version=__VERSION__,
    author='Dalibo',
    license='Postgresql',
    packages=find_packages(),
    install_requires=requires,
    description="A User Interface for the PoWA project",
    long_description="See http://dalibo.github.io/powa",
    scripts=['powa-web'],
    classifiers=[
    "Development Status :: 4 - Beta",
    "Intended Audience :: System Administrators",
    "License :: Other/Proprietary License",
    "License :: OSI Approved :: BSD License",
    "Operating System :: OS Independent",
    "Programming Language :: Python :: 2",
    "Programming Language :: Python :: 3",
    "Topic :: Database :: Front-Ends"],
    package_data={
        'powa': [
            'templates/*.html',
            'static/css/*',
            'static/js/powa.min-all.js',
            'static/js/config.js',
            'static/js/require.js',
            'static/img/**/*',
        ]
    }
)



"""
Utilities for commonly used SQL constructs.
"""
import re
from sqlalchemy.sql import text, select, func, case, column, extract, cast, bindparam
from sqlalchemy.types import Numeric
from collections import namedtuple

TOTAL_MEASURE_INTERVAL = """
extract( epoch from
    CASE WHEN min(total_mesure_interval) = '0 second'
        THEN '1 second'::interval
    ELSE min(total_mesure_interval) END)
"""


def format_jumbled_query(sql, params):
    it = iter(params)
    try:
        sql = re.sub("\?", lambda val: next(it), sql)
    except StopIteration:
        pass
    return sql


RESOLVE_OPNAME = text("""
    SELECT json_object_agg(oid, value)
    FROM (
    SELECT pg_operator.oid, json_build_object(
        'name', oprname,
        'indexams', array_agg(distinct pg_am.oid),
        'indexam_names', array_agg(distinct pg_am.amname)) as value
    FROM pg_operator
    JOIN pg_amop amop ON amop.amopopr = pg_operator.oid
    JOIN pg_am ON amop.amopmethod = pg_am.oid
    WHERE pg_operator.oid in :oid_list AND pg_am.amname != 'hash'
    GROUP BY pg_operator.oid, oprname
    ) detail
""")

RESOLVE_ATTNAME = text("""
    SELECT json_object_agg(attrelid || '.'|| attnum, value)
    FROM (
    SELECT attrelid, attnum, json_build_object(
        'relname', relname,
        'attname', attname,
        'n_distinct', stadistinct,
        'null_frac', stanullfrac,
        'most_common_values', CASE
            WHEN s.stakind1 = 1 THEN s.stavalues1
            WHEN s.stakind2 = 1 THEN s.stavalues2
            WHEN s.stakind3 = 1 THEN s.stavalues3
            WHEN s.stakind4 = 1 THEN s.stavalues4
            WHEN s.stakind5 = 1 THEN s.stavalues5
            ELSE NULL::anyarray
        END,
        'table_liverows', pg_stat_get_live_tuples(c.oid)
    ) as value
    FROM pg_attribute a
    INNER JOIN pg_class c on c.oid = a.attrelid
    LEFT JOIN pg_statistic s ON s.starelid = c.oid
                       AND s.staattnum = a.attnum
    WHERE (attrelid, attnum) IN :att_list
    ) detail
""")


def resolve_quals(conn, quallist, attribute="quals"):
    """
    Resolve quals definition (as dictionary coming from a to_json(quals)
    sql query.

    Arguments:
        conn: a connection to the database against which the qual was executed
        quallist: an iterable of rows, each storing quals in the attributes
        attribute: the attribute containing the qual list itself in each row
    """
    operator_to_look = set()
    attname_to_look = set()
    operators = {}
    attnames = {}
    for row in quallist:
        values = row[attribute]
        if not isinstance(values, list):
            values = [values]
        for v in values:
            operator_to_look.add(v['opno'])
            attname_to_look.add((v["relid"], v["attnum"]))
    if operator_to_look:
        operators = conn.execute(
            RESOLVE_OPNAME,
            {"oid_list": tuple(operator_to_look)}).scalar()
    if attname_to_look:
        attnames = conn.execute(
            RESOLVE_ATTNAME,
            {"att_list": tuple(attname_to_look)}).scalar()
    new_qual_list = []
    for row in quallist:
        row = dict(row)
        new_qual_list.append(row)
        values = row[attribute]
        if not isinstance(values, list):
            values = [values]
        for v in values:
            attname = attnames["%s.%s" % (v["relid"], v["attnum"])]
            v["relname"] = attname['relname']
            v["attname"] = attname['attname']
            v["n_distinct"] = attname['n_distinct']
            v["most_common_values"] = attname['most_common_values']
            v["null_frac"] = attname['null_frac']
            v["opname"] = operators[v["opno"]]['name']
            v["indexams"] = operators[v["opno"]]['indexams']
            v["indexam_names"] = operators[v["opno"]]['indexam_names']
            v["table_liverows"] = attname["table_liverows"]
            v["repr"] = "%s.%s %s ?" % (attname["relname"], attname["attname"],
                                        v["opname"])
        row["where_clause"] = "WHERE %s" % (" AND ".join(v["repr"] for v in values))
        row["eval_type"] = ", ".join("by index" if v["eval_type"] == 'i' else 'post-scan'
                                     for v in values)
    return new_qual_list


Plan = namedtuple(
    "Plan",
    ("title", "values", "query", "plan", "filter_ratio", "exec_count"))


def qual_constants(type, filter_clause, top=1):
    orders = {
        'most_executed': "4 DESC",
        'least_filtering': "6",
        'most_filtering': "6 DESC"
    }
    if type not in ('most_executed', 'most_filtering',
                    'least_filtering'):
        return
    filter_clause = filter_clause.compile()
    base = text("""
    (
    WITH sample AS (
    SELECT query, quals as quals,
                constants,
                sum(count) as count,
                sum(nbfiltered) as nbfiltered,
                CASE WHEN sum(count) = 0 THEN 0 ELSE sum(nbfiltered) / sum(count) END AS filter_ratio
        FROM powa_statements s
        JOIN pg_database ON pg_database.oid = s.dbid
        JOIN powa_qualstats_quals qn ON s.queryid = qn.queryid
        JOIN (
            SELECT *
            FROM powa_qualstats_constvalues_history qnc
            UNION ALL
            SELECT *
            FROM powa_qualstats_aggregate_constvalues_current
        ) qnc ON qn.qualid = qnc.qualid AND qn.queryid = qnc.queryid,
        LATERAL
                unnest(%s) as t(constants,nbfiltered,count)
        WHERE %s
        GROUP BY quals, constants, query
        ORDER BY %s
        LIMIT :top_value
    )
    SELECT query, quals, constants as constants, nbfiltered as nbfiltered,
                count as count,
                filter_ratio as filter_ratio,
                row_number() OVER (ORDER BY count desc NULLS LAST) as rownumber
        FROM sample
    ORDER BY 7
    LIMIT :top_value
    ) %s
    """ % (type, filter_clause.statement, orders[type], type)
                ).params(top_value=top, **filter_clause.params)
    return select(["*"]).select_from(base)



"""
Dashboard for the by-query page.
"""

from tornado.web import HTTPError
from sqlalchemy.sql import (
    bindparam, text, column, select, table,
    literal_column, case, cast, ColumnCollection,
    func, extract)
from sqlalchemy.sql.functions import sum
from sqlalchemy.types import Numeric

from powa.dashboards import (
    Dashboard, Graph, Grid,
    MetricGroupDef, MetricDef,
    DashboardPage, ContentWidget)
from powa.database import DatabaseOverview

from powa.sql import (Plan, format_jumbled_query,
                      resolve_quals, qual_constants)
from powa.sql.views import (powa_getstatdata_sample,
                            kcache_getstatdata_sample,
                            powa_getstatdata_detailed_db,
                            qualstat_getstatdata,
                            possible_indexes)
from powa.sql.utils import block_size, mulblock, greatest, to_epoch, total_measure_interval
from powa.sql.tables import powa_statements

class QueryOverviewMetricGroup(MetricGroupDef):
    """
    Metric Group for the graphs on the by query page.
    """
    name = "query_overview"
    xaxis = "ts"
    data_url = r"/metrics/database/(\w+)/query/(\w+)"
    rows = MetricDef(label="#Rows")
    calls = MetricDef(label="#Calls")
    shared_blks_read = MetricDef(label="Shared read", type="sizerate")
    shared_blks_hit = MetricDef(label="Shared hit", type="sizerate")
    shared_blks_dirtied = MetricDef(label="Shared dirtied", type="sizerate")
    shared_blks_written = MetricDef(label="Shared written", type="sizerate")
    local_blks_read = MetricDef(label="Local read", type="sizerate")
    local_blks_hit = MetricDef(label="Local hit", type="sizerate")
    local_blks_dirtied = MetricDef(label="Local dirtied", type="sizerate")
    local_blks_written = MetricDef(label="Local written", type="sizerate")
    temp_blks_read = MetricDef(label="Temp read", type="sizerate")
    temp_blks_written = MetricDef(label="Temp written", type="sizerate")
    blk_read_time = MetricDef(label="Read time", type="duration")
    blk_write_time = MetricDef(label="Write time", type="duration")
    avg_runtime = MetricDef(label="Avg runtime", type="duration")
    reads = MetricDef(label="Physical read", type="sizerate")
    writes = MetricDef(label="Physical writes", type="sizerate")
    user_time = MetricDef(label="CPU user time", type="percent")
    system_time = MetricDef(label="CPU system time", type="percent")
    hit_ratio = MetricDef(label="Shared buffers hit ratio", type="percent")
    miss_ratio = MetricDef(label="Shared buffers miss ratio", type="percent")
    sys_hit_ratio = MetricDef(label="System cache hit ratio", type="percent")
    disk_hit_ratio = MetricDef(label="Disk hit ratio", type="percent")

    @classmethod
    def _get_metrics(cls, handler, **params):
        base = cls.metrics.copy()
        if not handler.has_extension("pg_stat_kcache"):
            for key in ("reads", "writes", "user_time", "system_time",
                        "sys_hit_ratio", "disk_hit_ratio"):
                base.pop(key)
        else:
            base.pop("miss_ratio")

        return base

    @property
    def query(self):
        query = powa_getstatdata_sample("query")
        query = query.where(
            (column("datname") == bindparam("database")) &
            (column("queryid") == bindparam("query")))
        query = query.alias()
        c = query.c
        total_blocks = ((c.shared_blks_read + c.shared_blks_hit)
                        .label("total_blocks"))
        def bps(col):
            return (mulblock(col) / extract("epoch", greatest(c.mesure_interval, '1 second'))).label(col.name)
        cols = [to_epoch(c.ts),
                c.rows,
                c.calls,
                case([(total_blocks == 0, 0)],
                     else_=cast(c.shared_blks_hit, Numeric) * 100 / total_blocks
                     ).label("hit_ratio"),
                bps(c.shared_blks_read),
                bps(c.shared_blks_hit),
                bps(c.shared_blks_dirtied),
                bps(c.shared_blks_written),
                bps(c.local_blks_read),
                bps(c.local_blks_hit),
                bps(c.local_blks_dirtied),
                bps(c.local_blks_written),
                bps(c.temp_blks_read),
                bps(c.temp_blks_written),
                c.blk_read_time,
                c.blk_write_time,
                (c.runtime / greatest(c.calls, 1)).label("avg_runtime")]

        from_clause = query
        if self.has_extension("pg_stat_kcache"):
            # Add system metrics from pg_stat_kcache,
            # and detailed hit ratio.
            kcache_query = kcache_getstatdata_sample()
            kc = ColumnCollection(*kcache_query.inner_columns)
            kcache_query = (
                kcache_query
                .where(kc.queryid == bindparam("query"))
                .alias())
            kc = kcache_query.c
            sys_hits = (greatest(mulblock(c.shared_blks_read) -
                              kc.reads, 0)
                        .label("kcache_hitblocks"))
            sys_hitratio = (cast(sys_hits, Numeric) * 100 /
                            mulblock(total_blocks))
            disk_hit_ratio = (kc.reads  /
                              mulblock(total_blocks))
            cols.extend([
                kc.reads,
                kc.writes,
                kc.user_time,
                kc.system_time,
                case([(total_blocks == 0, 0)],
                     else_=disk_hit_ratio).label("disk_hit_ratio"),
                case([(total_blocks == 0, 0)],
                     else_=sys_hitratio).label("sys_hit_ratio")])
            from_clause = from_clause.join(
                kcache_query,
                kcache_query.c.ts == c.ts)
        else:
            cols.extend([
                case([(total_blocks == 0, 0)],
                     else_=cast(c.shared_blks_read, Numeric) * 100 / total_blocks
                     ).label("miss_ratio")
            ])

        return (select(cols)
                .select_from(from_clause)
                .where(c.calls != None)
                .order_by(c.ts)
                .params(samples=100))


class QueryIndexes(ContentWidget):
    """
    Content widget showing index creation suggestion.
    """

    data_url = r"/metrics/database/(\w+)/query/(\w+)/indexes"
    title = "Query Indexes"

    def get(self, database, query):
        if not self.has_extension("pg_qualstats"):
            raise HTTPError(501, "PG qualstats is not installed")
        base_query = qualstat_getstatdata()
        c = ColumnCollection(*base_query.inner_columns)
        base_query.append_from(text("""LATERAL unnest(quals) as qual"""))
        base_query = (base_query
                    .where(c.queryid == query)
                    .having(func.bool_or(column('eval_type') == 'f'))
                    .having(c.count > 1000)
                    .having(c.filter_ratio > 0.5)
                    .params(**{'from': '-infinity',
                                'to': 'infinity'}))
        optimizable = list(self.execute(base_query, params={'query': query}))
        optimizable = resolve_quals(self.connect(database=database),
                                    optimizable,
                                    'quals')
        qual_indexes = {}
        for line in optimizable:
            qual_indexes[line['where_clause']] = possible_indexes(line['quals'])
        self.render("database/query/indexes.html", indexes=qual_indexes)


class QueryExplains(ContentWidget):
    """
    Content widget showing explain plans for various const values.
    """
    title = "Query Explains"
    data_url = r"/metrics/database/(\w+)/query/(\w+)/explains"

    def get(self, database, query):
        if not self.has_extension("pg_qualstats"):
            raise HTTPError(501, "PG qualstats is not installed")
        condition = text("""datname = :database AND s.queryid = :query
                 AND coalesce_range && tstzrange(:from, :to)""")
        sql = (select(['most_filtering.quals',
                      'most_filtering.query',
                      'to_json(most_filtering) as "most filtering"',
                      'to_json(least_filtering) as "least filtering"',
                      'to_json(most_executed) as "most executed"'])
               .select_from(qual_constants("most_filtering", condition).alias("most_filtering")
               .join(qual_constants("least_filtering", condition).alias("least_filtering"),
                     text("most_filtering.rownumber = least_filtering.rownumber"))
               .join(qual_constants("most_executed", condition).alias("most_executed"),
                     text("most_executed.rownumber = least_filtering.rownumber"))))
        params = {"database": database, "query": query,
                  "from": self.get_argument("from"),
                  "to": self.get_argument("to")}
        quals = self.execute(sql, params=params)
        plans = []
        if quals.rowcount > 0:

            row = quals.first()
            for key in ('most filtering', 'least filtering', 'most executed'):
                vals = row[key]
                query = format_jumbled_query(row['query'], vals['constants'])
                plan = "N/A"
                try:
                    result = self.execute("EXPLAIN %s" % query,
                                          database=database)
                    plan = "\n".join(v[0] for v in result)
                except:
                    pass
                plans.append(Plan(key, vals['constants'], query,
                                  plan, vals["filter_ratio"], vals['count']))
        if len(plans) == 0:
            self.flash("No quals found for this query", "warning")
            self.render("xhr.html", content="")
            return

        self.render("database/query/explains.html", plans=plans)


class QualList(MetricGroupDef):
    """
    Datasource used for the Qual table.
    """
    name = "query_quals"
    xaxis = "quals"
    axis_type = "category"
    data_url = r"/metrics/database/(\w+)/query/(\w+)/quals"
    filter_ratio = MetricDef(label="Avg filter_ratio (excluding index)")
    count = MetricDef(label="Execution count (excluding index)")

    def prepare(self):
        if not self.has_extension("pg_qualstats"):
            raise HTTPError(501, "PG qualstats is not installed")

    @property
    def query(self):
        base = qualstat_getstatdata()
        c = ColumnCollection(*base.inner_columns)
        return (base.where(c.queryid == bindparam("query")))


    def process(self, val, database=None, query=None, **kwargs):
        row = dict(val)
        row['url'] = self.reverse_url('QualOverview', database, query,
                                      row['qualid'])
        return row

    def post_process(self, data, database, query, **kwargs):
        conn = self.connect(database=database)
        data["data"] = resolve_quals(conn, data["data"])
        return data


class QueryDetail(ContentWidget):
    """
    Detail widget showing summarized information for the query.
    """
    title = "Query Detail"
    data_url = r"/metrics/database/(\w+)/query/(\w+)/detail"

    def get(self, database, query):
        bs = block_size.c.block_size
        stmt = powa_getstatdata_detailed_db()
        stmt = stmt.where(
            (column("datname") == bindparam("database")) &
            (column("queryid") == bindparam("query")))
        stmt = stmt.alias()
        from_clause = stmt.join(powa_statements,
                                powa_statements.c.queryid == stmt.c.queryid)
        c = stmt.c
        rblk = mulblock(sum(c.shared_blks_read).label("shared_blks_read"))
        wblk = mulblock(sum(c.shared_blks_hit).label("shared_blks_hit"))
        stmt = (select([
            column("query"),
            sum(c.calls).label("calls"),
            sum(c.runtime).label("runtime"),
            rblk,
            wblk,
            (rblk + wblk).label("total_blks")])
            .select_from(from_clause)
            .group_by(column("query"), bs))

        value = self.execute(stmt, params={
            "query": query,
            "database": database,
            "from": self.get_argument("from"),
            "to": self.get_argument("to")
        })
        if value.rowcount < 1:
            self.render("xhr.html", content="No data")
            return
        self.render("database/query/detail.html", stats=value.first())


class QueryOverview(DashboardPage):
    """
    Dashboard page for a query.
    """
    base_url = r"/database/(\w+)/query/(\w+)/overview"
    params = ["database", "query"]
    datasources = [QueryOverviewMetricGroup, QueryDetail,
                   QueryExplains, QueryIndexes, QualList]
    parent = DatabaseOverview

    def __init__(self, *args, **kwargs):
        self._dashboard = None
        super(QueryOverview, self).__init__(*args, **kwargs)

    @classmethod
    def get_menutitle(cls, handler, params):
        return "Query detail"

    @property
    def dashboard(self):
        if self._dashboard:
            return self._dashboard
        hit_ratio_graph = Graph("Hit ratio",
                                metrics=[QueryOverviewMetricGroup.hit_ratio],
                                renderer="bar",
                                stack=True)
        self._dashboard = Dashboard(
            "Query %(query)s on database %(database)s",
            [[QueryDetail],
             [Graph("General",
                    metrics=[QueryOverviewMetricGroup.avg_runtime,
                             QueryOverviewMetricGroup.rows,
                             QueryOverviewMetricGroup.calls]),
              hit_ratio_graph],
             [Graph("Shared block (in Bps)",
                    metrics=[QueryOverviewMetricGroup.shared_blks_read,
                             QueryOverviewMetricGroup.shared_blks_hit,
                             QueryOverviewMetricGroup.shared_blks_dirtied,
                             QueryOverviewMetricGroup.shared_blks_written]),
              Graph("Local block (in Bps)",
                    metrics=[QueryOverviewMetricGroup.local_blks_read,
                             QueryOverviewMetricGroup.local_blks_hit,
                             QueryOverviewMetricGroup.local_blks_dirtied,
                             QueryOverviewMetricGroup.local_blks_written]),
              Graph("Temp block (in Bps)",
                    metrics=[QueryOverviewMetricGroup.temp_blks_read,
                             QueryOverviewMetricGroup.temp_blks_written])],
             [Graph("Read / Write time",
                    metrics=[QueryOverviewMetricGroup.blk_read_time,
                             QueryOverviewMetricGroup.blk_write_time])]])
        if self.has_extension("pg_stat_kcache"):
            self._dashboard.widgets.extend([[
                Graph("Physical block (in Bps)",
                      metrics=[QueryOverviewMetricGroup.reads,
                               QueryOverviewMetricGroup.writes]),
                Graph("CPU time",
                      metrics=[QueryOverviewMetricGroup.user_time,
                               QueryOverviewMetricGroup.system_time])]])
            hit_ratio_graph.metrics.append(QueryOverviewMetricGroup.sys_hit_ratio)
            hit_ratio_graph.metrics.append(QueryOverviewMetricGroup.disk_hit_ratio)
        else:
            hit_ratio_graph.metrics.append(QueryOverviewMetricGroup.miss_ratio)

        if self.has_extension("pg_qualstats"):
            self._dashboard.widgets.extend([[
                Grid("Predicates used by this query",
                     columns=[{
                         "name": "where_clause",
                         "label": "Predicate",
                         "type": "query",
                         "max_length": 60,
                         "url_attr": "url"
                     }, {
                         "name": "eval_type",
                         "label": "Eval Type"
                     }],
                     metrics=QualList.all())],
                [QueryIndexes],
                [QueryExplains]])
        return self._dashboard



from powa.framework import BaseHandler


class LoginHandler(BaseHandler):

    def get(self):
        self._status_code = 403
        return self.render("login.html", title="Login")

    def post(self, *args, **lwargs):
        username = self.get_argument('username')
        password = self.get_argument('password')
        server = self.get_argument('server')
        try:
            self.connect(username=username, password=password, server=server)
        except Exception as e:
            self.flash("Auth failed", "alert")
            self.get()
            return
        self.set_secure_cookie('username', username)
        self.set_secure_cookie('password', password)
        self.set_secure_cookie('server', server)
        self.redirect(self.get_argument('next', '/'))


class LogoutHandler(BaseHandler):

    def get(self):
        self.clear_all_cookies()
        return self.redirect('/')
